BasicError

Enum BasicError 

pub enum BasicError {
    NotFound(String),
    InvalidArgument(String),
    NotAllowed(String),
    PermissionDenied(String),
    Internal(String),
}
A basic error that occurs while processing a WAMP message.

Variants



NotFound(String)

A generic resource was not found.

WAMP defines standard URIs for not finding specific resource types. This error should only be used when the standard URI cannot be used.



InvalidArgument(String)

An invalid argument was passed.



NotAllowed(String)

The operation is not allowed based on process configuration.



PermissionDenied(String)

The operation is not allowed based on user permissions.



Internal(String)

Some internal error occurred.

Should only be used when there is no other error variant that describes the error, since the message is very vague and not very useful for debugging.

Implementations

Source

impl BasicError

pub fn uri_component(&self) -> &str

The trailing URI component for the error.